package org.bitbucket.bradleysmithllc.etlunit.feature.file;

import org.bitbucket.bradleysmithllc.etlunit.Log;
import org.bitbucket.bradleysmithllc.etlunit.RuntimeSupport;
import org.bitbucket.bradleysmithllc.etlunit.TestExecutionError;
import org.bitbucket.bradleysmithllc.etlunit.feature.file.json.file._assert.AssertRequest;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.io.file.DataFileManager;
import org.bitbucket.bradleysmithllc.etlunit.io.file.DataFileSchema;
import org.bitbucket.bradleysmithllc.json.validator.ResourceNotFoundException;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObject;
import org.bitbucket.bradleysmithllc.etlunit.util.IOUtils;

import javax.inject.Inject;
import javax.inject.Named;
import java.io.File;
import java.io.FileFilter;
import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

public class FileRuntimeSupportImpl implements FileRuntimeSupport
{
	public static final String DEFAULT = "default";
	private RuntimeSupport runtimeSupport;
	private Map<String, FileProducer> fileProducerMap = new HashMap<String, FileProducer>();

	private DataFileManager dataFileManager;
	private Log applicationLog;

	public FileRuntimeSupportImpl()
	{
		System.out.println();
	}

	@Inject
	public void receiveApplicationLog(@Named("applicationLog") Log log)
	{
		applicationLog = log;
	}

	@Inject
	public void receiveRuntimeSupport(RuntimeSupport runtime)
	{
		runtimeSupport = runtime;
	}

	@Inject
	public void receiveDataFileManager(DataFileManager dfm)
	{
		dataFileManager = dfm;
	}

	@Override
	public File getDataFileForCurrentTest(String name)
	{
		return getDataFile(runtimeSupport.getCurrentlyProcessingTestPackage(), name);
	}

	public File getDataFile(String package_, String name)
	{
		return new File(getDataFileDir(package_), name);
	}

	public File getDataFile(String name)
	{
		return getDataFile(null, name);
	}

	@Override
	public File getAssertionFileForCurrentTest(String name)
	{
		return getAssertionFile(runtimeSupport.getCurrentlyProcessingTestPackage(), name);
	}

	@Override
	public File getAssertionFileDirForCurrentTest() {
		return getAssertionFileDir(runtimeSupport.getCurrentlyProcessingTestPackage());
	}

	public File getAssertionFileDir(String package_)
	{
		return new FileBuilder(runtimeSupport.getTestResourceDirectory(package_, "data")).mkdirs().file();
	}

	public File getAssertionFileDir()
	{
		return getDataFileDir(null);
	}

	public File getAssertionFile(String package_, String name)
	{
		return new FileBuilder(getAssertionFileDir(package_)).mkdirs().name(name + ".delimited").file();
	}

	public File getAssertionFile(String name)
	{
		return getAssertionFile(null, name);
	}

	@Override
	public File getGeneratedDataFileForCurrentTest(String name)
	{
		return getGeneratedDataFile(runtimeSupport.getCurrentlyProcessingTestPackage(), name);
	}

	@Override
	public File getGeneratedFileSchemaFileForCurrentTest(String name) {
		return getGeneratedFileSchemaFile(runtimeSupport.getCurrentlyProcessingTestPackage(), name);
	}

	public File getGeneratedDataFile(String package_, String name)
	{
		FileBuilder fb = new FileBuilder(runtimeSupport.getGeneratedSourceDirectory("file")).subdir("generated-data");

		if (package_ != null)
		{
			fb = fb.subdir(package_);
		}

		fb.name(name);

		return fb.mkdirsToFile().file();
	}

	@Override
	public File getGeneratedFileSchemaFile(String _package, String name) {
		FileBuilder fb = new FileBuilder(runtimeSupport.getGeneratedSourceDirectory("file")).subdir("fml");

		if (_package != null)
		{
			fb = fb.subdir(_package);
		}

		fb.name(name + ".fml");

		return fb.mkdirsToFile().file();
	}

	public File getGeneratedDataFile(String name)
	{
		return getGeneratedDataFile(null, name);
	}

	public File getReferenceFile(String path, String name)
	{
		return runtimeSupport.getReferenceFile(path, name);
	}

	@Override
	public File getDataFileDirForCurrentTest()
	{
		return getDataFileDir(runtimeSupport.getCurrentlyProcessingTestPackage());
	}

	public File getReferenceFileSchema(String package_, String name)
	{
		FileBuilder fb = new FileBuilder(runtimeSupport.getReferenceDirectory("file/fml"));

		if (package_ != null)
		{
			fb = fb.subdir(package_).mkdirs();
		}

		File sch = fb.name(name + ".fml").file();

		return sch;
	}

	public File getReferenceFileSchema(String name)
	{
		return getReferenceFileSchema(null, name);
	}

	@Override
	public DataFileSchema locateReferenceFileSchemaForCurrentTest(String name, String generifiedName, ETLTestValueObject parameters) throws RequestedFmlNotFoundException, ResourceNotFoundException {
		return locateReferenceFileSchema(runtimeSupport.getCurrentlyProcessingTestPackage(), name, generifiedName, parameters);
	}

	public DataFileSchema locateReferenceFileSchema(String package_, String name, String generifiedName, ETLTestValueObject parameters) throws RequestedFmlNotFoundException, ResourceNotFoundException
	{
		applicationLog.info("Searching for an fml");

		if (parameters != null)
		{
			ETLTestValueObject query = parameters.query("reference-file-type");
			if (query != null)
			{
				String referenceId = query.getValueAsString();
				File referenceFileSchema = getReferenceFileSchema(package_, referenceId);

				if (referenceFileSchema.exists())
				{
					applicationLog.info("Resolved to explicit reference file schema: " + referenceFileSchema.getAbsolutePath());
					return dataFileManager.loadDataFileSchema(referenceFileSchema);
				}
				else
				{
					// this is an error if the explicitly-defined type is not found
					String resourcePath = "reference/file/fml/" + referenceId + ".fml";
					applicationLog.info("Searching classpath for explicit reference file schema: " + resourcePath);

					try
					{
						return dataFileManager.loadDataFileSchemaFromResource(resourcePath);
					}
					catch(ResourceNotFoundException exc)
					{
						throw new RequestedFmlNotFoundException("reference-file-type [" + referenceId + "] not found", exc);
					}
				}
			}
		}

		if (name != null)
		{
			File referenceFileSchema = getReferenceFileSchema(package_, name);

			if (referenceFileSchema.exists())
			{
				applicationLog.info("Resolved to implicitly-matched reference file schema: " + referenceFileSchema.getAbsolutePath());
				return dataFileManager.loadDataFileSchema(referenceFileSchema);
			}
			else
			{
				try
				{
					String resourcePath = "reference/file/fml/" + name + ".fml";
					applicationLog.info("Searching classpath for explicit reference file schema: " + resourcePath);
					return dataFileManager.loadDataFileSchemaFromResource(resourcePath);
				}
				catch(ResourceNotFoundException rnfe)
				{
					// just continue on - this is normal
				}
			}
		}

		File referenceFileSchema = getReferenceFileSchema(package_, generifiedName);

		if (referenceFileSchema.exists())
		{
			applicationLog.info("Resolved to generified reference file schema: " + referenceFileSchema.getAbsolutePath());
			return dataFileManager.loadDataFileSchema(referenceFileSchema);
		}
		else
		{
			String resourcePath = generifiedName + ".fml";
			applicationLog.info("Searching classpath for generified reference file schema: " + resourcePath);
			return dataFileManager.loadDataFileSchemaFromResource(resourcePath);
		}
	}

	public DataFileSchema locateReferenceFileSchema(String name, String generifiedName, ETLTestValueObject parameters) throws RequestedFmlNotFoundException, ResourceNotFoundException {
		return locateReferenceFileSchema(null, name, generifiedName, parameters);
	}

	public void persistDataFileSchema(DataFileSchema schema) throws IOException
	{
		persistDataFileSchema(null, schema);
	}

	public void persistDataFileSchema(String package_, DataFileSchema schema) throws IOException
	{
		File file = getReferenceFileSchema(package_, schema.getId());

		IOUtils.writeBufferToFile(file, new StringBuffer(schema.toJsonString()));
	}

	public void persistGeneratedDataFileSchema(DataFileSchema schema) throws IOException
	{
		persistGeneratedDataFileSchema(null, schema);
	}

	public void persistGeneratedDataFileSchema(String package_, DataFileSchema schema) throws IOException
	{
		File file = getGeneratedFileSchemaFile(package_, schema.getId());

		IOUtils.writeBufferToFile(file, new StringBuffer(schema.toJsonString()));
	}

	@Override
	public void persistGeneratedDataFileSchemaForCurrentTest(DataFileSchema schema) throws IOException {
		persistGeneratedDataFileSchema(runtimeSupport.getCurrentlyProcessingTestPackage(), schema);
	}

	@Override
	public void persistDataFileSchemaForCurrentTest(DataFileSchema schema) throws IOException
	{
		persistDataFileSchema(runtimeSupport.getCurrentlyProcessingTestPackage(), schema);
	}

	public File getDataFileDir(String package_)
	{
		return new FileBuilder(runtimeSupport.getTestResourceDirectory(package_, "files")).mkdirs().file();
	}

	public File getDataFileDir()
	{
		return getDataFileDir(null);
	}

	public void registerFileProducer(FileProducer producer)
	{
		if (fileProducerMap.containsKey(producer.getName()))
		{
			throw new IllegalArgumentException("File producer [" + producer.getName() + "] already registered as class [" + fileProducerMap.get(producer.getName()).getClass() + "]");
		}

		if (fileProducerMap.size() == 0)
		{
			fileProducerMap.put(DEFAULT, producer);
		}

		fileProducerMap.put(producer.getName(), producer);
	}

	public FileProducer getRegisteredProducer(String name)
	{
		if (name == null)
		{
			name = DEFAULT;
		}

		if (!fileProducerMap.containsKey(name))
		{
			throw new IllegalArgumentException("File producer [" + name + "] not registered.");
		}

		return fileProducerMap.get(name);
	}

	public List<String> getReferenceFileSchemaPackages()
	{
		return runtimeSupport.getReferencePackages("file/fml");
	}

	public List<String> getReferenceFileSchemasForPackage(String package_)
	{
		File fmlRef = runtimeSupport.getReferenceDirectory(package_ == null ? "file/fml" : ("file/fml/" + package_));

		final List<String> schemas = new ArrayList<String>();

		fmlRef.listFiles(new FileFilter()
		{
			public boolean accept(File pathname)
			{
				if (pathname.isFile() && IOUtils.fileHasExtension(pathname, "fml"))
				{
					schemas.add(IOUtils.removeExtension(pathname));
				}

				return false;
			}
		});

		return schemas;
	}

	public List<File> getReferenceFileSchemaFilesForPackage(String package_)
	{
		File fmlRef = runtimeSupport.getReferenceDirectory(package_ == null ? "file/fml" : ("file/fml/" + package_));

		final List<File> schemas = new ArrayList<File>();

		fmlRef.listFiles(new FileFilter()
		{
			public boolean accept(File pathname)
			{
			if (pathname.isFile() && IOUtils.fileHasExtension(pathname, "fml"))
			{
				schemas.add(pathname);
			}

			return false;
			}
		});

		return schemas;
	}

	@Override
	public List<String> getReferenceFileSchemasForPackageForCurrentTest()
	{
		return getReferenceFileSchemasForPackage(runtimeSupport.getCurrentlyProcessingTestPackage());
	}

	@Override
	public File getReferenceFileSchemaForCurrentTest(String name)
	{
		return getReferenceFileSchema(runtimeSupport.getCurrentlyProcessingTestPackage(), name);
	}
}